#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""Module for session handling and encapsulation"""

##############################################################################
#    Copyright (C) 2012-2013 by Javier Peña Blazquez <javipenya@gmail.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
##############################################################################

##############################################################################
# @author: Javier Peña Blázquez
# @copyright: 2012-2013 Javier Peña Blazquez
# @license: http://www.apache.org/licenses/LICENSE-2.0  Apache License 2.0
# @contact: <javipenya@gmail.com>
# @version: 0.1
##############################################################################

import sys
from Session import Session
from EasyPyPoAError import SessionError

class Singleton:
    """
    A non-thread-safe helper class to ease implementing singletons.
    This should be used as a decorator -- not a metaclass -- to the
    class that should be a singleton.

    The decorated class can define one `__init__` function that
    takes only the `self` argument. Other than that, there are
    no restrictions that apply to the decorated class.

    To get the singleton instance, use the `Instance` method. Trying
    to use `__call__` will result in a `TypeError` being raised.

    Limitations: The decorated class cannot be inherited from.
    """

    def __init__(self, decorated):
        self._decorated = decorated

    def Instance(self):
        """
        Returns the singleton instance. Upon its first call, it creates a
        new instance of the decorated class and calls its `__init__` method.
        On all subsequent calls, the already created instance is returned.
        """
        try:
            return self._instance
        except AttributeError:
            self._instance = self._decorated()
            return self._instance

    def __call__(self):
        raise TypeError('Singletons must be accessed through `Instance()`.')

    def __instancecheck__(self, inst):
        return isinstance(inst, self._decorated)

@Singleton
class AuthSession(object):
    """Manages and encapsulation of a Python session"""
    def __init__(self):
        self._session = None
        self._serviceID = None
        
    def start(self, serviceID):
        self._serviceID = serviceID
        try:
            self._session = Session(serviceID)
            self._session.start()
        except Exception as e:
            message = "Failed to create PAPI session"
            raise SessionError(message, e)
    
    def destroy(self, locationTo=None):
        try:
            self._session.destroy()
        except Exception as e:
            message = "Failed to destroy PAPI session"
            raise SessionError(message, e)
        if locationTo != None:
            print 'Location: ' + locationTo + '\n'
            sys.exit()
    
    def commit(self):
        self._session.commit()
    
    def output(self):
        return self._session.output()
        
    def get(self, key):
        if self._session.isset(key):
            return self._session[key]
        return False
    
    def has(self, key):
        return self._session.isset(key)
        
    def set(self, key, value):
        self._session[key] = value
    
    def get_serviceID(self):
        return self._serviceID
    
    def has_serviceID(self):
        return self._serviceID != None
        
    def get_authZ(self):
        return self.get('AUTHZ')
    
    def has_authZ(self):
        return self.has('AUTHZ')
    
    def set_authZ(self, authZ):
        self.set('AUTHZ', authZ)
        
    def get_userdata(self):
        return self.get('USERDATA')
    
    def has_userdata(self):
        return self.has('USERDATA')
    
    def set_userdata(self, userdata):
        self.set('USERDATA', userdata)
        
    def get_issuer(self):
        return self.get('ISSUER')
    
    def has_issuer(self):
        return self.has('ISSUER')
    
    def set_issuer(self, issuer):
        self.set('ISSUER', issuer)
        
    def get_timestamp(self):
        return self.get('TIMESTAMP')
    
    def has_timestamp(self):
        return self.has('TIMESTAMP')
    
    def set_timestamp(self, timestamp):
        self.set('TIMESTAMP', timestamp)
    
    def get_requestURL(self):
        return self.get('REQUESTURL')
    
    def has_requestURL(self):
        return self.has('REQUESTURL')
    
    def set_requestURL(self, requestURL):
        self.set('REQUESTURL', requestURL)
        
    def get_key(self):
        return self.get('KEY')
    
    def has_key(self):
        return self.has('KEY')
    
    def set_key(self, key):
        self.set('KEY', key)
