#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""Module for cryptographic functions"""

##############################################################################
#    Copyright (C) 2013 by Javier Peña Blazquez
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
##############################################################################

##############################################################################
# @author: Javier Peña Blázquez
# @copyright: 2012-2013 Javier Peña Blazquez
# @license: http://www.apache.org/licenses/LICENSE-2.0  Apache License 2.0
# @contact: <javipenya@gmail.com>
# @version: 0.1
##############################################################################

import M2Crypto
import base64
import math
from EasyPyPoAError import CryptError

class Crypt(object):
    """Class for decrypt GPoA messages

    Attributes:
        pubKey -- GPoA public Key
        pubKeySize -- Bits of the public key 
    """
    def __init__(self, publicKeyFile):
        """Constructor 
        Gets GPoA public key path. Load public key with M2Crypto module and 
        calculates the number of bits of the key.

        publicKeyFile String: GPoA public key path
        """
        try:
            self._pubKey = M2Crypto.RSA.load_pub_key(publicKeyFile)
            self._pubKeySize = len(self._pubKey)
        except Exception as e:
            message = "Cannot read GPoA public key: "
            raise CryptError(message, e)
    
    def decrypt(self, message):
        """
        Decrypts the received message with the GPoA public key.
        
        message String: encrypted and base64 encoded message
        """
        # Decode the base64 input string
        try:
            cryptData = base64.b64decode(message)
            #cryptData = base64.urlsafe_b64decode(message)
        except Exception as e:
            message = "Can not base64 decode response"
            raise CryptError(message, e)
        
        # Get the byte size of encrypted data 
        dataSize = len(cryptData)
        
        # Get the input block maximum size in Bytes
        inputBlockSize = self._pubKeySize/8
        numBlocks = math.ceil(dataSize/float(inputBlockSize))
        
        # Start to decrypt
        blockCount = 0;
        decryptBuffer = []
        
        while blockCount < numBlocks:
            index = blockCount * inputBlockSize
            block = cryptData[index:index+inputBlockSize]
            # Decrypt the text
            try:
                decrypttext = self._pubKey.public_decrypt(block, M2Crypto.RSA.pkcs1_padding)
            except M2Crypto.RSA.RSAError as e:
                message = "Can not decrypt response, check GPoA public key"
                raise CryptError(message, e)
            decryptBuffer.append(decrypttext)
            blockCount+=1
        
        # Now join the array with the blocks string encripted
        decryptData = ''.join(decryptBuffer)
        # Return the base64 decode, decrypted and joined data string
        return decryptData
   
