#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""Implements the PoA component of a PAPI architecture"""

##############################################################################
#    Copyright (C) 2012-2013 by Javier Peña Blazquez <javipenya@gmail.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
##############################################################################

##############################################################################
# @author: Javier Peña Blázquez
# @copyright: 2012-2013 Javier Peña Blazquez
# @license: http://www.apache.org/licenses/LICENSE-2.0  Apache License 2.0
# @contact: <javipenya@gmail.com>
# @version: 0.1
##############################################################################

import os
import re
import logging
import cgi
import time
from EasyPyPoAError import PoAError
from PoAConfig import PoAConfig
from AuthSession import AuthSession
from Message import (CheckMessage, CheckedMessage, PapiSignOffReqMessage, 
                     PapiLoggedOutMessage, PapiLogOutMessage)


class EasyPyPoA(object):
    
    def __init__(self, serviceID, configFile=None):
        self._config = PoAConfig(serviceID, configFile)
        self._serviceID = serviceID
        self._session = None
        self._requestURL = self._get_requestURL()
        self._initialize_log_system(self._config.get_log_level(), 
                                    self._config.get_log_file())
        logging.info("PoA started successfully for serviceID [" + 
                     self._serviceID + "] and Location [" + 
                     self._config.get_location() + "]")
        
    def get_serviceID(self):
        return self._serviceID
    
    def check_access(self):
        if self._session == None:
            self._session = AuthSession.Instance()
            self._session.start(self._serviceID)
        data = cgi.FieldStorage()
        action = data.getfirst('ACTION')
        if action != None:
            # Process response from a GPoA
            action = cgi.escape(action)
            if action == 'CHECKED':
                logging.info("Processing CHECKED message")
                checked = CheckedMessage(self._session, data, self._config)
                return checked.process()
                
            elif action == 'PAPILOGGEDOUT':
                logging.info("Processing PAPILOGGEDOUT message")
                papiloggedout = PapiLoggedOutMessage(self._session, data, 
                                                     self._config)
                papiloggedout.process()
            
            elif action == 'PAPILOGOUT':
                if (self._session.get_serviceID() == self._serviceID and 
                    self._session.get_authZ() == True):
                    logging.info("Processing PAPILOGOUT message")
                    papilogout = PapiLogOutMessage(self._session, data, 
                                                   self._config)
                    papilogout.process()
                else:
                    message = ("Received a PAPILOGOUT message without "
                               "previous session" )
                    logging.error(message)
                    raise PoAError(message)
                    
            else:
                message = ("The value [" + action + "] is not valid for "
                           "the ACTION field.")
                logging.error(message)
                raise PoAError(message)
        elif (self._session.get_serviceID() == self._serviceID and 
              self._session.get_authZ() == True):
            logging.info("Authenticated user, checking token validity")
            return self._is_authorized()
        else:
            logging.info("Unauthenticated user, sending CHECK request to GPoA")
            check = CheckMessage(self._session, self._requestURL, 
                                 self._config.get_GPoA_URL())
            check.send()
        return False
    
    def sign_off_request(self):
        if self._session == None:
            self._session = AuthSession.Instance()
            self._session.start(self._serviceID)
        if (self._session.get_serviceID() != self._serviceID or 
            self._session.get_authZ() != True):
            message = "Can not send a Logout if not set before a session"
            logging.error(message)
            raise PoAError(message)
        papisignoffreq = PapiSignOffReqMessage(self._session, 
                                               self._config.get_GPoA_URL())
        papisignoffreq.send()
        
          
    def get_user_data(self):
        if self._session == None:
            message = ("Can not get user data if not set before a session")
            raise PoAError(message)
        else:
            if self._session.has_userdata():
                return self._session.get_userdata()
            logging.warning("No details of the user's assertion")
            return {}
    
    def get_issuer(self):
        if self._session == None:
            message = ("Can not get issuer if not set before a session")
            raise PoAError(message)
        else:
            if self._session.has_issuer():
                return self._session.get_issuer()
            logging.warning("No details of the issuer of the assertion")
            return ""
    
    def set_cookie(self):
        if self._session != None:
            return self._session.output()
        return ''
    
    def _is_authorized(self):
        """Checks if the user has a session that authorizes access.
        Return True if successful access or make a CHECK request to GPoA
        in other case
        """
        
        self._session.set_authZ(False)
        self._session.commit()
        timestamp = self._session.get_timestamp()
        ctime = int(time.time())
        timeout = self._config.get_lcook_timeout()
        if timestamp + timeout < ctime:
            logging.info("Expired session, new request to GPoA")
            logging.debug("Current Time: " + str(ctime))
            logging.debug("Session TimeStamp: " + str(timestamp))
            logging.debug("Session TimeOut: " + str(timeout))
            check = CheckMessage(self._session, self._requestURL, 
                                 self._config.get_GPoA_URL())
            check.send()
            
        # Regenerate the session
        logging.info("Regenerating the user session")
        self._session.set_timestamp(time.time())
        self._session.set_authZ(True)
        self._session.commit()
        return True
        
    
    def _initialize_log_system(self, loglevel, logfile):
        _loglevel = getattr(logging, loglevel, None)
        log_format = '%(levelname)s: %(asctime)s %(message)s'
        date_format = '[%Y/%m/%d] %H:%M:%S'
        try:
            logging.basicConfig(filename=logfile, level=_loglevel, 
                                format=log_format, datefmt=date_format)
        except Exception as e:
            message = "Fail to initialize logging system."
            raise PoAError(message, e)
        
    def _get_requestURL(self):
        if os.environ.has_key('HTTPS') and os.environ.get('HTTPS') == 'on':
            schema = 'https://'
        else:
            schema = 'http://'
        host = os.environ.get('HTTP_HOST', '_EXCEPTION_')
        request = os.environ.get('REQUEST_URI', '_EXCEPTION_')
        
        requestURL = schema + host + request 
        pattern = '_EXCEPTION_'
        
        if re.search(pattern, requestURL) != None:
            message = ("Can not compose the URL of this PoA [" + 
                       requestURL + "]")
            raise PoAError(message)
        else:
            return requestURL
            
