#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""Module for handling PAPI messages"""

##############################################################################
#    Copyright (C) 2012-2013 by Javier Peña Blazquez <javipenya@gmail.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
##############################################################################

##############################################################################
# @author: Javier Peña Blázquez
# @copyright: 2012-2013 Javier Peña Blazquez
# @license: http://www.apache.org/licenses/LICENSE-2.0  Apache License 2.0
# @contact: <javipenya@gmail.com>
# @version: 0.1
##############################################################################

import logging
import random
import hashlib
import urllib
import os
import time
from HTTPManager import HTTPManager
from EasyPyPoAError import PoAError
from Crypt import Crypt

class Message(object):
    """Base class for messages in EasyPyPoA module."""
    pass

class CheckMessage(Message):
    """Responsible for generating CHECK messages"""
    
    def __init__(self, session, requestURL, GPoAURL):
        self._session = session
        self._requestURL = requestURL
        self._GPoAURL = GPoAURL
    
    def send(self):
        """Send a CHECK request to GPoA/IdP for checking user data."""
        
        # Generates alphanumeric string that should be returned in the response
        key = hashlib.sha1(str(random.randint(999, 9999) * 
                               random.randint(999, 9999))).hexdigest()
        logging.debug("Generating random token key {%s} to keep the request" % key)
        
        # Save session data
        self._save_session(key, self._requestURL)
        
        # Generates URL for the CHECK message
        http = self._generate_URL(self._GPoAURL, key, self._requestURL)
        
        # Redirect 302 to GPoA URL
        http.HTTP_redirect(self._session)
        
    def _save_session(self, key, requestURL):
        self._session.set_authZ('False')
        self._session.set_key(key)
        self._session.set_requestURL(requestURL)
        self._session.commit()
        
    def _generate_URL(self, GPoAURL, key, requestURL):
        http = HTTPManager(GPoAURL)
        http.clean_parameters()
        http.add_parameter('ACTION', 'CHECK')
        http.add_parameter('DATA', key)
        http.add_parameter('URL', requestURL)
        logging.debug("Request CHECK for GPoA: [%s]" % str(http))
        return http
    

class CheckedMessage(Message):
    """Responsible for process CHECKED messages"""
    
    GPOA_PUBKEY_FILE = '_GPoA_pubkey.pem'
    ISSUER_ID_SEPARATOR  = '@'
    
    def __init__(self, session, data, config):
        self._session = session
        self._data = data
        self._config = config

    def process(self):
        """Responsible for processing the CHECKED response received from the 
        GPoA"""
        # Get local time
        ctime = int(time.time())
        # Get DATA parameter
        DATA = self._get_DATA()
        # Decrypt DATA
        decryptedDATA = self._decrypt_DATA(DATA)
        # Split DATA
        splitedDATA = self._split_DATA(decryptedDATA)
        
        # Extract data fields from the assertion
        key = splitedDATA.pop()
        try:
            currentTime = int(splitedDATA.pop())
            expiryTime = int(splitedDATA.pop())
        except ValueError as e:
            message = ("Incorrect value for currentTime or expiryTime in "
                       "CHECKED response. Must be integer values. ") 
            raise PoAError(message, e)
        userAssertion =  ':'.join(splitedDATA)
        logging.debug("KEY: " + key + ", CurrentTime: " + str(currentTime) +
                      ", ExpiryTime: " + str(expiryTime) + ", UserAssertion: " +
                      userAssertion)
        
        # Check if the userAssertion DATA field is ERROR
        if userAssertion == 'ERROR':
            # Two possibilities:
            # 1. - The user is not authorized to access PoA.
            # 2. - Indicate has happened some kind of error when identifying.
            message = ("The user is not authorized to access the PoA "
                       "or an error has occurred in the identity provider "
                       "when identifying the user")
            logging.warning(message)
            return False
        
        # Verify expiryTime
        self._verify_expiryTime(expiryTime, ctime)
        # Verify currentTime
        self._verify_currentTime(currentTime, ctime)
        # Verify KEY
        self._verify_key(key)
        
        # Split userAssertion
        splitedUserAssert = self._split_userAssertion(userAssertion)
        
        # Extract data fields from the UserAssertion
        # ID of the IdP CHECKED message sender
        issuer = splitedUserAssert.pop()
        self._verify_issuer(issuer)
        logging.debug("Issuer ID: " + issuer)
        
        # User Data (list of attributes)
        userDataField = CheckedMessage.ISSUER_ID_SEPARATOR.join(splitedUserAssert)
        logging.debug("UserData: " + userDataField)
        
        # Gets a dictionary with the attributes and their values
        userData = self._process_UserData(userDataField)
        
        # Save data session
        self._save_session(issuer, userData)
        
        logging.info("Access permission granted to resource " + 
                     self._session.get_requestURL())
        
        return True    
    
    def _get_DATA(self):
        data = self._data.getfirst('DATA')
        if data != None:
            return urllib.unquote(data)
        else:
            message = ("Incorrect GPoA CHECKED response. No DATA param")
            logging.error(message)
            raise PoAError(message)
    
    def _decrypt_DATA(self, data):
        pubkeypath = os.path.join(self._config.get_pubkey_path(), 
                                  CheckedMessage.GPOA_PUBKEY_FILE)
        cipher = Crypt(pubkeypath)
        decryptedDATA = cipher.decrypt(data)
        logging.debug("Decrypted DATA: " + decryptedDATA)
        return decryptedDATA
        
    def _split_DATA(self, data):
        splitedDATA = data.split(':')
        if len(splitedDATA) < 4:
            message = ("Incorrect GPoA CHECKED response. DATA parameter "
                       "format is not correct. DATA => [" + data + "]")
            raise PoAError(message)
        return splitedDATA
    
    def _verify_expiryTime(self, expiryTime, ctime):
        if expiryTime <= ctime:
            message = ("Incorrect GPoA CHECKED response. Received assertion "
                       "has expired")
            logging.error(message)
            raise PoAError(message)
    
    def _verify_currentTime(self, currentTime, ctime):
        # Debug log messages
        logging.debug("Local time: [%s]" % str(ctime))
        logging.debug("Assertion Time Expedition: [%s]" % str(currentTime))
        logging.debug("Assertion Time Expedition - windows: [%s]" 
                      % str(currentTime - self._config.get_current_time_windows()))
        logging.debug("Assertion Time Expedition + windows: [%s]" 
                      % str(currentTime + self._config.get_current_time_windows()))
        
        # Verifies that currentTime precedes the local time
        # considering the time offset that may exist between the systems
        if ctime < (currentTime - self._config.get_current_time_windows()):
            message = ("Incorrect GPoA CHECKED response. Assertion generated "
                       "before being requested. This error is due to a "
                       "desynchronization between IdP, GPoA and PoA")
            logging.error(message)
            raise PoAError(message)
        # Verifies the URL Timeout for the CHECKED response 
        elif ctime > (currentTime + self._config.get_url_timeout()):
            message = ("Incorrect GPoA CHECKED response. Assertion expired. "
                       "Has been too long since the GPoA generated the "
                       "response for this PoA")
            logging.error(message)
            raise PoAError(message)
    
    def _verify_key(self, key):
        # Check if the response of GPoA corresponds 
        # to a CHECK message of this PoA
        if (not self._session.has_key()) or (self._session.get_key() != key):
            message = ("Incorrect GPoA CHECKED response. Response not "
                       "correspond to any previous CHECK request of this PoA")
            logging.error(message)
            logging.debug("CHECKED KEY: %s" % key)
            logging.debug("Session KEY: %s" % self._session.get_key())
            raise PoAError(message)
        
    def _split_userAssertion(self, userAssertion):
        splitedUserAssert = userAssertion.split(CheckedMessage.ISSUER_ID_SEPARATOR)
        if len(splitedUserAssert) < 2:
            message = ("Incorrect GPoA CHECKED response. UserAssertion "
                       "format is not correct. Found a problem getting the "
                       "ID of the issuer of the assertion.\n")
            message += ("UserAssertion: " + userAssertion + "\n")
            message += ("Issuer_ID_Separator: " + CheckedMessage.ISSUER_ID_SEPARATOR)
            logging.error(message)
            raise PoAError(message)
        return splitedUserAssert
    
    def _verify_issuer(self, issuer):
        # Check that Issuer identifier does not have tokens used to separate 
        # attribute-value pairs, attribute and value or values, ie '=', and 
        # generally ',' and '|'.
        if (issuer.count(self._config.get_attributes_separator()) > 0 or
            issuer.count(self._config.get_values_separator()) > 0 or
            issuer.count('=') > 0): 
            message = ("The ID of the IdP message sender is not in a valid "
                       "format")
            raise PoAError(message)
    
    def _process_UserData(self, userData):
        # Dictionary to store the attributes and their values
        userDataDict = {}
        if userData != '':
            # The assertion contains user data. Process it.
            # Split attributes
            attributes = userData.split(self._config.get_attributes_separator())
            for attr in attributes:
                # Separate the attribute name and values
                splitedAttr = attr.split('=')
                # Verify that two components are obtained: attrName and attrValue
                if len(splitedAttr) != 2:
                    message = ("Problem getting the attrName and the "
                               "attrValue of the attribute: " + attr)
                    logging.error(message)
                    raise PoAError(message)
                
                # Save the attribute name
                attrName = splitedAttr[0]
                if attrName == '':
                    message = ("Problem processing assertion: key/value "
                               "pair " + attr + " with empty or undefined key")
                    logging.warning(message)
                # Save the attribute values
                attrValues = splitedAttr[1]
                if attrValues == '':
                    message = ("Problem processing assertion: key/value "
                               "pair " + attr + " with empty or undefined value")
                    logging.warning(message)
                    
                # Separate the different values ​​of the attributes
                splitedValues = attrValues.split(self._config.get_values_separator())
                logging.debug("Pair <key, value/s> = <" + attrName + ", " + attrValues + ">")
                
                # Save attributes and values in dictionary
                userDataDict[attrName] = splitedValues
                    
        return userDataDict
    
    def _save_session(self, issuer, userData):
        self._session.set_issuer(issuer)
        self._session.set_userdata(userData)
        self._session.set_timestamp(int(time.time()))
        self._session.set_authZ(True)
        self._session.commit()


class PapiSignOffReqMessage(Message):
    """Responsible for destroy PAPI session and 
    generating PAPISIGNOFFREQ messages 
    """
    
    def __init__(self, session, GPoAURL):
        self._session = session
        self._GPoAURL = GPoAURL
        if not self._session.has_requestURL():
            message = ("On request PAPISIGNOFFREQ not URL parameter defined")
            raise PoAError(message)
        self._requestURL = self._session.get_requestURL()
        self._serviceID = self._session.get_serviceID()
            
    
    def send(self):
        """Destroy PAPI Session and send a PAPISIGNOFFREQ request to 
        GPoA/IdP"""
        
        # Generates URL for the PAPISIGNOFFREQ message
        http = self._generate_URL(self._GPoAURL, self._requestURL, 
                                  self._serviceID)
        
        # Destroy PAPI Session
        self._session.destroy()
        
        # Redirect 302 to GPoA URL
        logging.info("Starting Logout")
        http.HTTP_redirect()
        
        
    def _generate_URL(self, GPoAURL, requestURL, serviceID):
        http = HTTPManager(GPoAURL)
        http.clean_parameters()
        http.add_parameter('ACTION', 'PAPISIGNOFFREQ')
        http.add_parameter('DATA', 'DUMMY')
        http.add_parameter('URL', requestURL)
        http.add_parameter('POA', serviceID)
        http.add_parameter('PAPIOPOA', requestURL)
        logging.debug("Request PAPISIGNOFFREQ for GPoA: [%s]" % str(http))
        return http
    
    
class PapiLoggedOutMessage(Message):
    """Responsible for process PAPILOGGEDOUT messages"""
    
    def __init__(self, session, data, config):
        self._session = session
        self._data = data
        self._config = config
        
    def process(self):
        # Get DATA parameter
        DATA = self._data.getfirst('DATA')
        # Get URL (optional) parameter
        URL = self._data.getfirst('URL')
        # Get PAPIOPOA (optional) parameter
        PAPIOPOA = self._data.getfirst('PAPIOPOA')
        
        # Verify DATA
        self._verify_DATA(DATA)
        
        # Verify the URL of where comes this response
        self._verify_URL(URL, self._config.get_GPoA_URL())
        
        # Verify PAPIOPOA
        self._verify_PAPIOPOA(PAPIOPOA)
        
        # Destroy PAPI Session
        logging.info("Redirect the user to the Logout URL specified "
                     "in the PoA configuration")
        logging.debug("Logout URL: [" + self._config.get_end_logout() + "]")
        self._session.destroy(self._config.get_end_logout())
        
    def send(self, requestURL, PAPIOPOA=None):
        # Generates URL for the PAPILOGGEDOUT message
        http = self._generate_URL(self._config.get_GPoA_URL(), 
                                  requestURL, PAPIOPOA)
        
        # Redirect 302 to GPoA URL
        http.HTTP_redirect()
        
    def _generate_URL(self, GPoAURL, requestURL, PAPIOPOA):
        http = HTTPManager(GPoAURL)
        http.clean_parameters()
        http.add_parameter('ACTION', 'PAPILOGGEDOUT')
        http.add_parameter('DATA', 'DUMMY')
        http.add_parameter('URL',  requestURL)
        if PAPIOPOA != None:
            http.add_parameter('PAPIOPOA', PAPIOPOA)
        logging.debug("Request PAPILOGGEDOUT for GPoA: [%s]" % str(http))
        return http
        
    def _verify_DATA(self, DATA):
        if DATA == None:
            message = ("Incorrect GPoA PAPILOGGEDOUT response. No DATA param")
            logging.error(message)
            raise PoAError(message)
        
    def _verify_URL(self, URL, GPoAURL):
        # If URL parameter exists
        if URL != None:
            if urllib.unquote(URL) != GPoAURL:
                # The response comes from an unknown GPoA
                message = ("The PAPILOGGEDOUT response received does not "
                           "come from GPoA which was sent ")
                message += ("Sent to: [" + GPoAURL + "]\n")
                message += ("Received from: [" + URL + "]\n")
                logging.error(message)
                raise PoAError(message)
            else:
                logging.debug("Response PAPILOGGEDOUT comes from GPoA with "
                              "URL [" + GPoAURL + "]")
                
    def _verify_PAPIOPOA(self, PAPIOPOA):
        if PAPIOPOA != None:
            logging.debug("LOGOUT Request began in: " + urllib.unquote(PAPIOPOA))
                
    
class PapiLogOutMessage(Message):
    """Responsible for process PAPILOGOUT messages"""
    
    def __init__(self, session, data, config):
        self._data = data
        self._config = config
        self._session = session
        
    def process(self):
        # Get DATA parameter
        DATA = self._data.getfirst('DATA')
        # Get URL parameter
        URL = self._data.getfirst('URL')
        # Get PAPIOPOA (optional) parameter
        PAPIOPOA = self._data.getfirst('PAPIOPOA')
        
        # Verify DATA
        self._verify_DATA(DATA)
        
        # Verify the URL of where comes this request
        self._verify_URL(URL, self._config.get_GPoA_URL())
        
        # Verify PAPIOPOA
        self._verify_PAPIOPOA(PAPIOPOA)
        
        # Save request URL
        requestURL = self._session.get_requestURL()
        
        # Destroy PAPI Session
        logging.info("Closed PAPI session by a GPoA request")
        self._session.destroy()
        
        # Run user-provided code to logout of the application behind the PoA
        if self._config.get_hook_logout() != None:
            func = self._config.get_hook_logout()
            try:
                eval(func)()
            except Exception:
                message = ("An error has occurred by executing the "
                           "function/method defined by the user in " 
                           "Hook_Logout")
                logging.warning(message)
        
        # Reply to closing session
        papiloggedout = PapiLoggedOutMessage(self._session, self._data, 
                                             self._config)
        papiloggedout.send(requestURL, PAPIOPOA)

        
    def _verify_DATA(self, DATA):
        if DATA == None:
            message = ("Incorrect GPoA PAPILOGOUT request. No DATA param")
            logging.error(message)
            raise PoAError(message)  
    
    def _verify_URL(self, URL, GPoAURL):
        # If URL parameter exists
        if URL != None:
            if urllib.unquote(URL) != GPoAURL:
                # The request comes from an unknown GPoA
                message = ("The PAPILOGOUT request does not "
                           "come from known GPoA ")
                message += ("Received from: [" + URL + "]\n")
                message += ("Expected from: [" + GPoAURL + "]\n")
                logging.error(message)
                raise PoAError(message)
            else:
                logging.debug("Request PAPILOGOUT comes from GPoA with "
                              "URL [" + GPoAURL + "]")
        else:
            message = ("Incorrect GPoA PAPILOGOUT request. No URL param")
            logging.error(message)
            raise PoAError(message)
        
    def _verify_PAPIOPOA(self, PAPIOPOA):
        if PAPIOPOA != None:
            logging.debug("LOGOUT Request began in: " + urllib.unquote(PAPIOPOA))
            
            
