#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""Module for PoA Configuration"""

##############################################################################
#    Copyright (C) 2012-2013 by Javier Peña Blazquez <javipenya@gmail.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
##############################################################################

##############################################################################
# @author: Javier Peña Blázquez
# @copyright: 2012-2013 Javier Peña Blazquez
# @license: http://www.apache.org/licenses/LICENSE-2.0  Apache License 2.0
# @contact: <javipenya@gmail.com>
# @version: 0.1
##############################################################################

import ConfigParser
import os
from EasyPyPoAError import ConfigurationError

class PoAConfig(object):
    """Class representing the configuration of the PoA.

    Attributes:
        _lcookTimeout -- Time during which no checking of credentials is 
                         performed. Similar to the lifetime of the session.
        _urlTimeout -- Lifetime of response messages issued by a GPoA.
        _currentTimewindows -- Time window that allows interoperability 
                               between systems with time delay.
        _attributesSeparator -- Token used to separate the pairs 
                                Attr_name = Attr_values in assertions.
        _valuesSeparator -- Token used to separate the multiple values that
                            can have an attribute in assertions.
        _logLevel -- Level of the log messages generated by EasyPyPoA.
        _logFile -- Path to the file where the log messages are sent.
        _pubkeyPath -- Path to the directory where is saved the file with 
                       the public key of GPoA. The file must be named 
                       _GPoA_Pubkey.pem.
        _GPoAUrl -- GPoA URL which will be issued messages and from which 
                    the replies are received.
        _location -- Path of the URL that will be protected by the PoA.
        _serviceID -- PoA identifier for which the configuration is processed.
        _endLogout -- URL where user will be redirected after making a LOGOUT
        _hookLogout -- Method or user-defined function that will be invoked 
                       by EasyPyPoA when logout is performed to allow the 
                       Relying Party (the application that relies on EasyPoA) 
                       make also logout.
    """
    
    # Default configuration file path
    EASYPYPOA_CONFIG_FILE = os.path.join(os.path.dirname(__file__), 
                                         'conf', 'PAPIPoA.ini')
    
    def __init__(self, serviceID, configFile):
        """Constructor 
        Gets the configuration of a PoA from serviceID and configuration file. 
        If not receive any configuration file, seek one in the conf/ directory located 
        in the same directory where the file with EasyPyPoA class.

        serviceID String: service identifier
        configFile String: configuration file path
        """
        
        # Default values of data attributes
        self._lcookTimeout = 3600
        self._urlTimeout = 10
        self._currentTimeWindows = 2
        self._attributesSepatator = ','
        self._valuesSeparator = '|'
        self._logLevel = 'WARNING'
        self._logFile = os.path.join(os.path.dirname(__file__), 
                                  'logs', 'PAPI_EasyPyPoA.log')
        self._pubKeyPath = os.path.join(os.path.dirname(__file__), 
                                  'conf', 'keys', '')
        
        self._serviceID = serviceID
        self._endLogout = None
        self._hookLogout = None
        
        if not configFile:
            configFile = PoAConfig.EASYPYPOA_CONFIG_FILE
        config = self._load_config_file(configFile)
        self._parse_config_file(config)
        
    def get_lcook_timeout(self):
        """Gets the configuration directive Lcook_Timeout
        Return: Seconds of the session lifetime
        """
        return self._lcookTimeout
    
    def get_url_timeout(self):
        """Gets the configuration directive URL_Timeout
        Return: Seconds for a valid CHECKED message
        """
        return self._urlTimeout
    
    def get_current_time_windows(self):
        """Gets the configuration directive Current_Time_Windows
        Return: Seconds of the time window allowed desynchronization 
        between the GPoA and the PoA
        """
        return self._currentTimeWindows
    
    def get_attributes_separator(self):
        """Gets the configuration directive Attributes_Separator
        Return: character used to separate the pairs (attr_name, attr_values) 
        in PAPI assertions.
        """
        return self._attributesSepatator
    
    def get_values_separator(self):
        """Gets the configuration directive Values_Separator
        Return: character used to separate the multiple values that can have 
        an attribute in PAPI assertions.
        """
        return self._valuesSeparator
    
    def get_log_level(self):
        """Gets the configuration directive LogLevel
        Return: a log level defined in logging module
        """
        return self._logLevel
    
    def get_log_file(self):
        """Gets the configuration directive LogFile
        Return: path to EasyPyPoA log file
        """
        return self._logFile
    
    def get_pubkey_path(self):
        """Gets the configuration directive Pubkeys_Path
        Return: path to the directory where is saved the public key of GPoA
        """
        return self._pubKeyPath
    
    def get_GPoA_URL(self):
        """Gets the configuration directive GPoA_URL
        Return: GPoA URL
        """
        return self._GPoAUrl
    
    def get_location(self):
        """Gets the configuration directive Location for instantiated PoA
        Return: URL path protected by the PoA
        """
        return self._location
    
    def get_serviceID(self):
        """Gets the PoA identifier
        Return: PoA identifier
        """
        return self._serviceID
    
    def get_end_logout(self):
        """Gets URL which redirects the user to the end of the 
        logout process
        Return: end Logout URL
        """
        return self._endLogout
    
    def get_hook_logout(self):
        """Gets the name of the method or function
        Return: function name
        """
        return self._hookLogout
    
    def _load_config_file(self, configFile):
        """Verify that the configuration file exists and is readable.
        Uses the ConfigParser module to verify that the file has the 
        structure of an INI file.
        
        Arguments:
        configFile -- configuration file path
        
        Return: ConfigParser class instance with the configuration read
        """
        config = ConfigParser.ConfigParser()
        try:
            with open(configFile):
                try:
                    config.read(configFile)
                except (ConfigParser.ParsingError, 
                        ConfigParser.MissingSectionHeaderError) as e:
                    message = "Configuration file has incorrect format"
                    raise ConfigurationError(message, e)
        except IOError as e:
            message = "Configuration file can't be accessed"
            raise ConfigurationError(message, e)
        return config
    
    def _parse_config_file(self, config):
        """Verify that the configuration file has a section with the 
        service ID, and the options GPoA_URL and Location.
        Load all the options values. 
        
        Arguments:
        config -- ConfigParser instance with the configuration read 
        """
        if not config.has_section(self._serviceID):
            message = ("Configuration file has no settings for PoA with "
                       "the service identifier: [" + self._serviceID + "]")
            raise ConfigurationError(message)
        
        required_options = ['GPoA_URL', 'Location']
        for option in required_options:
            if not config.has_option(self._serviceID, option):
                message = ("Configuration file has no " + option + " "
                           "directive for PoA with the service identifier" 
                           ": [" + self._serviceID + "]")
                raise ConfigurationError(message)
        
        # Load all options values
        self._load_gpoa_url(config)
        self._load_location(config)
        self._load_lcook_timeout(config)
        self._load_url_timeout(config)
        self._load_current_time_windows(config)
        self._load_attributes_separator(config)
        self._load_values_separator(config)
        self._load_log_level(config)
        self._load_log_file(config)
        self._load_pubkeys_path(config)
        self._load_end_logout_url(config)
        self._load_hook_logout(config)

    def _load_gpoa_url(self, config):
        self._GPoAUrl = config.get(self._serviceID, 'GPoA_URL')
        
    def _load_location(self, config):
        self._location = config.get(self._serviceID, 'Location')
        if not self._valid_location(self._location):
            message = ("Trying to get credentials in an unprotected location "
                       "in the configuration file: " + 
                       os.environ.get('REQUEST_URI', ''))
            raise ConfigurationError(message)
        
    def _load_lcook_timeout(self, config):
        if config.has_option(self._serviceID, 'Lcook_Timeout'):
            try:
                self._lcookTimeout = config.getint(self._serviceID, 
                                                   'Lcook_Timeout')
            except ValueError as e:
                message = ("Invalid value for Lcook_Timeout directive in "
                           "configuration file.")
                raise ConfigurationError(message, e)
        
    def _load_url_timeout(self, config):
        if config.has_option(self._serviceID, 'URL_Timeout'):
            try:
                self._urlTimeout = config.getint(self._serviceID, 
                                                 'URL_Timeout')
            except ValueError as e:
                message = ("Invalid value for URL_Timeout directive in "
                       "configuration file.")
                raise ConfigurationError(message, e)
            
    def _load_current_time_windows(self, config):
        if config.has_option(self._serviceID, 'Current_Time_Windows'):
            try:
                self._currentTimeWindows = config.getint(self._serviceID,
                                                         'Current_Time_Windows')
            except ValueError as e:
                message = ("Invalid value for Current_Time_Windows directive "
                           "in configuration file.")
                raise ConfigurationError(message, e)
        
    def _load_attributes_separator(self, config):
        if config.has_option(self._serviceID, 'Attribute_Separator'):
            attr_sep = config.get(self._serviceID, 'Attribute_Separator')
            attr_sep = attr_sep.strip('"\' ')
            if len(attr_sep) == 1:
                self._attributesSepatator = attr_sep
            else:
                message = ("Invalid value for Attribute_Separator directive "
                           "in configuration file: " + attr_sep)
                raise ConfigurationError(message)
            
    def _load_values_separator(self, config):
        if config.has_option(self._serviceID, 'Value_Separator'):
            value_sep = config.get(self._serviceID, 'Value_Separator')
            value_sep = value_sep.strip('"\' ')
            if len(value_sep) == 1:
                self._valuesSeparator = value_sep
            else:
                message = ("Invalid value for Value_Separator directive in "
                           "configuration file: " + value_sep)
                raise ConfigurationError(message)
            
    def _load_log_level(self, config):
        if config.has_option(self._serviceID, 'LogLevel'):
            loglevel = config.get(self._serviceID, 
                                  'LogLevel').strip('"\' ').upper()
            if loglevel in ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']:
                self._logLevel = loglevel
            else:
                message = ("Invalid value for LogLevel directive in "
                           "configuration file: " + loglevel)
                raise ConfigurationError(message)
            
    def _load_log_file(self, config):
        if config.has_option(self._serviceID, 'LogFile'):
            logfile = config.get(self._serviceID, 'LogFile')
            try:
                with open(logfile, 'a'):
                    self._logFile = logfile
            except IOError as e:
                message = ("Invalid value for LogFile directive in "
                           "configuration file. Log file " + logfile + 
                           " can't be accessed.")
                raise ConfigurationError(message, e)
            
    def _load_pubkeys_path(self, config):
        if config.has_option(self._serviceID, 'Pubkeys_Path'):
            pubkeys_path = config.get(self._serviceID, 'Pubkeys_Path')
            if os.path.isdir(pubkeys_path):
                self._pubKeyPath = pubkeys_path
            else:
                message = ("Invalid value for Pubkeys_Path directive in "
                           "configuration file: " + pubkeys_path + " is not "
                           "a valid path.")
                raise ConfigurationError(message)
            
    def _load_end_logout_url(self, config):
        if config.has_option(self._serviceID, 'End_Logout'):
            self._endLogout = config.get(self._serviceID, 'End_Logout')
        else:
            if os.environ.has_key('HTTPS') and os.environ.get('HTTPS') == 'on':
                schema = 'https://'
            else:
                schema = 'http://'
            host_location = os.environ.get('HTTP_HOST') + '/' + self._location + '/?loggedout=true'
            host_location = host_location.replace('//', '/')
            self._endLogout = schema + host_location
            
    def _load_hook_logout(self, config):
        if config.has_option(self._serviceID, 'Hook_Logout'):
            self._hookLogout = config.get(self._serviceID, 'Hook_Logout')

    def _valid_location(self, location):
        if os.environ.has_key('REQUEST_URI'):
            web_path = os.environ.get('REQUEST_URI')
            return web_path[:len(location)] == location
        else:
            message = ("Unable to validate the URL protected by the PoA: "
                       "REQUEST_URI empty")
            raise ConfigurationError(message)
            
