#!/usr/bin/env python
# -*- coding: UTF-8 -*-

""" Module for session handling
    Note: A user must have cookies enabled.
"""

##############################################################################
#    Copyright (C) 2012-2013 by Javier Peña Blazquez <javipenya@gmail.com>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
##############################################################################

##############################################################################
# @author: Javier Peña Blázquez
# @copyright: 2012-2013 Javier Peña Blazquez
# @license: http://www.apache.org/licenses/LICENSE-2.0  Apache License 2.0
# @contact: <javipenya@gmail.com>
# @version: 0.1
##############################################################################

import Cookie
import cPickle
import time
import hashlib
import os
import datetime
from EasyPyPoAError import SessionError

# The following class is a modified version of this:
# http://www.sunjay.ca/download/session.py
# The original software is licensed under the MIT License.
# Copyright (c) 2010 - Sunjay Varma

class Session(object):
    """Class for session handling"""
    
    # Path for session files 
    # expanduser for windows users 
    SESSION_DIR = os.path.expanduser("~/.py_sessions") 
    SESSION_EXT = ".pys" # session file extension
    TODAY = str(datetime.date.today())
    
    def __init__(self, name="__sid__"):
        self._data = {}
        self._started = False
        self._name = name
        date = datetime.date.today() + datetime.timedelta(days=365)
        self._expires = date.strftime("%a, %d-%b-%Y %H:%M:%S GMT") # expires in one year
        #self._expires = 0 # delete right away
        if not os.environ.has_key('HTTP_COOKIE'):
            os.environ['HTTP_COOKIE'] = ""
        if not os.path.exists(Session.SESSION_DIR):
            os.makedirs(Session.SESSION_DIR)
        self._sid = self._getsid()
        self._path = os.path.join(Session.SESSION_DIR, self._sid 
                                  + Session.SESSION_EXT)
        
    def isset(self, name):
        """Is the variable set in the session?"""
        if not self._started:
            message = "Session must be started"
            raise SessionError(message)

        return name in self

    def unset(self, name):
        """Unset the name from the session"""
        if not self._started:
            message = "Session must be started"
            raise SessionError(message)
        del self[name]
        
    @staticmethod
    def _newsid():
        """Create a new session ID"""
        h = hashlib.sha224()
        h.update(str(time.time()/time.clock()**-1)+str(os.getpid()))
        return h.hexdigest()
        
    def _getsid(self):
        """Get the current session ID or return a new one"""
        # Try to load the sid from the HTTP cookie
        self._cookie = Cookie.SimpleCookie()
        if os.environ.has_key('HTTP_COOKIE'):
            self._cookie.load(os.environ['HTTP_COOKIE'])

            if self._name in self._cookie:
                sid = self._cookie[self._name].value
                return sid
            # return a new sid
            return self._newsid()
        else:
            message = "Could not find any cookies"
            raise SessionError(message)

    def start(self):
        """Start the session"""
        if self._started:
            return True # session cannot be started more than once per script

        # Load the session if it exists
        if os.path.exists(self._path):
            with open(self._path, "rb") as f:
                self._data = dict(cPickle.load(f))
                self._data["__date_loaded__"] = Session.TODAY
        else: # create a session
            with open(self._path, "wb") as f:
                self._data = {"__date_loaded__": Session.TODAY}

        # The session is officially started!
        self._started = True

        # Store the sid in the cookie
        self._cookie.clear()
        self._cookie[self._name] = self._sid
        self._cookie[self._name]['path'] = '/'
        self._cookie[self._name]['domain'] = os.environ['HTTP_HOST'].split(':')[0]
        self._cookie[self._name]['expires'] = str(self._expires)
        self._cookie[self._name]['httponly'] = True
        self._cookie[self._name]['version'] = '1'

        return True
    
    def commit(self):
        """Commit the changes to the session"""
        if not self._started:
            message = "Session must be started"
            raise SessionError(message)
        with open(self._path, "wb") as f:
            cPickle.dump(self._data, f, cPickle.HIGHEST_PROTOCOL)

    def destroy(self):
        """Destroy the session"""
        if not self._started:
            message = "Session must be started"
            raise SessionError(message)
        os.unlink(self._path)
        self._started = False
        return True

    def output(self):
        """Commit changes and send headers."""
        if not self._started:
            message = "Session must be started"
            raise SessionError(message)
        self.commit()
        return self._cookie.output()

    def __getitem__(self, item):
        """Get the item from the session"""
        if not self._started:
            message = "Session must be started"
            raise SessionError(message)
        return self._data.__getitem__(item)

    def __setitem__(self, item, value):
        """Set the item into the session"""
        if not self._started:
            message = "Session must be started"
            raise SessionError(message)
        self._data.__setitem__(item, value)

    def __delitem__(self, item):
        if not self._started:
            message = "Session must be started"
            raise SessionError(message)
        self._data.__delitem__(item)

    def __contains__(self, item):
        """Return if item in the session"""
        if not self._started:
            message = "Session must be started"
            raise SessionError(message)
        return self._data.__contains__(item)

    def __iter__(self):
        """Go through the names of all the session variables"""
        if not self._started:
            message = "Session must be started"
            raise SessionError(message)
        return self._data.__iter__()


